@@ -1,5 +1,5 @@
 #/*##########################################################################
-# Copyright (C) 2004-2020 V.A. Sole, ESRF - D. Dale CHESS
+# Copyright (C) 2004-2022 V.A. Sole, ESRF - D. Dale CHESS
 #
 # This file is part of the PyMca X-ray Fluorescence Toolkit developed at
 # the ESRF by the Software group.
@@ -31,17 +31,19 @@ __copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
 import sys
 import os
 import posixpath
+import time
 import gc
 import re
 from operator import itemgetter
 import logging
 _logger = logging.getLogger(__name__)
-try:
+
+if "hdf5plugin" not in sys.modules:
     # try to import hdf5plugins
-    import hdf5plugin
-except:
-    # but do not crash just because of it
-    pass
+    try:
+        import hdf5plugin
+    except:
+        _logger.info("Cannot import hdf5plugin")
 import h5py
 import weakref
 
@@ -73,14 +75,13 @@ else:
     def MyQVariant(x=None):
         return x
 
-
-
 QVERSION = qt.qVersion()
 
 
 #sorting method
-def h5py_sorting(object_list):
-    sorting_list = ['start_time', 'end_time', 'name']
+def h5py_sorting(object_list, sorting_list=None):
+    if sorting_list is None:
+        sorting_list = ['start_time', 'end_time', 'name']
     n = len(object_list)
     if n < 2:
         return object_list
@@ -116,6 +117,30 @@ def h5py_sorting(object_list):
             return object_list
 
     try:
+        if sorting_key == 'title':
+            def getTitle(x):
+                try:
+                    title = x["title"][()]
+                except:
+                    # allow the title to be missing
+                    title = ""
+                if hasattr(title, "dtype"):
+                    if hasattr(title, "__len__"):
+                        if len(title) == 1:
+                            title = title[0]
+                if hasattr(title, "decode"):
+                    title = title.decode("utf-8")
+                return title
+            # sort first by the traditional keys in order to be sorted
+            # by title and respecting actquisition order for equal title
+            try:
+                ordered_list = h5py_sorting(object_list)
+            except:
+                ordered_list = object_list
+            sorting_list = [(getTitle(o[1]), o) for o in ordered_list]
+            sorted_list = sorted(sorting_list, key=itemgetter(0))
+            return [x[1] for x in sorted_list]
+
         if sorting_key != 'name':
             sorting_list = [(o[1][sorting_key][()], o)
                            for o in object_list]
@@ -131,23 +156,22 @@ def h5py_sorting(object_list):
         #The only way to reach this point is to have different
         #structures among the different entries. In that case
         #defaults to the unfiltered case
-        print("WARNING: Default ordering")
-        print("Probably all entries do not have the key %s" % sorting_key)
+        _logger.warning("WARNING: Default ordering")
+        _logger.warning("Probably all entries do not have the key %s" % sorting_key)
         return object_list
 
+
 def _get_number_list(txt):
     rexpr = '[/a-zA-Z:_-]'
     nbs= [float(w) for w in re.split(rexpr, txt) if w not in ['',' ']]
     return nbs
 
+
 class BrokenLink(object):
     pass
 
+
 class QRLock(qt.QMutex):
-
-    """
-    """
-
     def __init__(self):
         qt.QMutex.__init__(self, qt.QMutex.Recursive)
 
@@ -203,25 +227,50 @@ class RootItem(object):
         del self._children[idx]
         del self._identifiers[idx]
 
+
 class H5NodeProxy(object):
+    def sortChildren(self, column, order):
+        #print("sort children called with ", column, order)
+        self.__sorting = True
+        if column == 1:
+            self.__sorting_list = ["title"]
+        else:
+            self.__sorting_list = None
+        self.__sorting_order = order
+
+    def raw_keys(self):
+        for node in self.children:
+            yield node.name
+
+    def raw_values(self):
+        node = self.getNode(self.name)
+        if hasattr(node, "values"):
+            return node.values()
+        else:
+            # spech5 does not implement values() prior to silx 0.6.0
+            for _, value in self.raw_items():
+                yield value
+
+    def raw_items(self):
+        return self.getNode(self.name).items()
+
     @property
     def children(self):
         if not self.hasChildren:
             return []
 
-        if not self._children:
+        if self.__sorting or not self._children:
             # obtaining the lock here is necessary, otherwise application can
             # freeze if navigating tree while data is processing
             if 1: #with self.file.plock:
-                items = list(self.getNode(self.name).items())
-                if posixpath.dirname(self.name) == "/":
-                    # top level item
-                    doit = True
-                else:
-                    doit = False
+                try:
+                    items = list(self.raw_items())
+                except Exception:
+                    items = []
+                    _logger.warning("Cannot obtain items list. Ignoring")
                 try:
                     # better handling of external links
-                    finalList = h5py_sorting(items)
+                    finalList = h5py_sorting(items, sorting_list=self.__sorting_list)
                     for i in range(len(finalList)):
                         if finalList[i][1] is not None:
                             finalList[i][1]._posixPath = posixpath.join(self.name,
@@ -233,22 +282,19 @@ class H5NodeProxy(object):
                                                                finalList[i][0])
                     self._children = [H5NodeProxy(self.file, i[1], self)
                                       for i in finalList]
-                except:
-                    #one cannot afford any error, so I revert to the old
+                except Exception:
+                    # one cannot afford any error, so I revert to the old
                     # method where values where used instead of items
-                    if _logger.getEffectiveLevel() == logging.DEBUG:
+                    if 1 or _logger.getEffectiveLevel() == logging.DEBUG:
                         raise
-                    else:
-                        # tmpList = list(self.getNode(self.name).values())
-                        # spech5 does not implement values() prior to silx 0.6.0
-                        tmpList = list(map(self.getNode(self.name).__getitem__,
-                                           self.getNode(self.name).keys()))
-                        finalList = tmpList
+                    tmpList = list(self.raw_values())
+                    finalList = tmpList
                     for i in range(len(finalList)):
                         finalList[i]._posixPath = posixpath.join(self.name,
                                                                items[i][0])
                     self._children = [H5NodeProxy(self.file, i, self)
                                       for i in finalList]
+        self.__sorting = False
         return self._children
 
     @property
@@ -278,6 +324,8 @@ class H5NodeProxy(object):
     def shape(self):
         if type(self._shape) == type(""):
             return self._shape
+        if self._shape is None:
+            return ""
         if len(self._shape) == 1:
             return "%d" % self._shape[0]
         elif len(self._shape) > 1:
@@ -305,6 +353,9 @@ class H5NodeProxy(object):
         return self._color
 
     def __init__(self, ffile, node, parent=None, path=None):
+        self.__sorting = False
+        self.__sorting_list = None
+        self.__sorting_order = qt.Qt.AscendingOrder
         if 1:#with ffile.plock:
             self._file = ffile
             self._parent = parent
@@ -397,6 +448,28 @@ class H5FileProxy(H5NodeProxy):
         else:
             return H5NodeProxy(self.file, self.file[path], self)
 
+    def raw_keys(self):
+        if isinstance(self.file, h5py.File):
+            file_path = self.file.filename
+            data_path = self.name
+            try:
+                from PyMca5.PyMcaIO import HDF5Utils
+                return HDF5Utils.safe_hdf5_group_keys(file_path,
+                                                      data_path=data_path)
+            except:
+                _logger.debug("Using standard approach")
+                return self.file[data_path].keys()
+        else:
+            return super().raw_keys()
+
+    def raw_values(self):
+        for _, value in self.raw_items():
+            yield value
+
+    def raw_items(self):
+        for data_path in self.raw_keys():
+            yield data_path, self.getNode(data_path)
+
 
 class FileModel(qt.QAbstractItemModel):
     """
@@ -409,6 +482,11 @@ class FileModel(qt.QAbstractItemModel):
         self.rootItem = RootItem(['File/Group/Dataset', 'Description', 'Shape', 'DType'])
         self._idMap = {qt.QModelIndex().internalId(): self.rootItem}
 
+    def sort(self, column, order):
+        #print("FileModel sort called with ", column, order)
+        for item in self.rootItem:
+            item.sortChildren(column, order)
+
     def clearRows(self, index):
         self.getProxyFromIndex(index).clearChildren()
 
@@ -421,52 +499,73 @@ class FileModel(qt.QAbstractItemModel):
         return 4
 
     def data(self, index, role):
-        if role == qt.Qt.DisplayRole:
-            item = self.getProxyFromIndex(index)
-            column = index.column()
-            if column == 0:
-                if isinstance(item, H5FileProxy):
-                    return MyQVariant(os.path.basename(item.file.filename))
-                else:
-                    if hasattr(item, "name"):
-                        return MyQVariant(posixpath.basename(item.name))
+        try:
+            if role == qt.Qt.DisplayRole:
+                item = self.getProxyFromIndex(index)
+                column = index.column()
+                if column == 0:
+                    if isinstance(item, H5FileProxy):
+                        try:
+                            return MyQVariant(os.path.basename(item.file.filename))
+                        except:
+                            _logger.critical("Cannot retrieve file name")
+                            return MyQVariant("Unknown file. Please refresh")
                     else:
-                        # this can only happen with the root
-                        return MyQVariant("/")
-            if column == 1:
-                showtitle = True
-                if showtitle:
-                    if hasattr(item, 'type'):
-                        if item.type in ["Entry", "NXentry"]:
-                            children = item.children
-                            names = [posixpath.basename(o.name) for o in children]
-                            if "title" in names:
-                                idx = names.index("title")
-                                node = children[idx].getNode()
-                                if hasattr(node, "shape") and len(node.shape):
-                                    #stored as an array of strings???
-                                    #return just the first item
-                                    return MyQVariant("%s" % node[()][0])
+                        if hasattr(item, "name"):
+                            return MyQVariant(posixpath.basename(item.name))
+                        else:
+                            # this can only happen with the root
+                            return MyQVariant("/")
+                if column == 1:
+                    showtitle = True
+                    if showtitle:
+                        if hasattr(item, 'type'):
+                            if item.type in ["Entry", "NXentry"]:
+                                if hasattr(item, "children"):
+                                    children = item.children
+                                    names = [posixpath.basename(o.name) for o in children]
+                                    if "title" in names:
+                                        idx = names.index("title")
+                                        node = children[idx].getNode()
+                                        if hasattr(node, "shape") and len(node.shape):
+                                            #stored as an array of strings???
+                                            #return just the first item
+                                            if hasattr(node, "asstr"):
+                                                try:
+                                                    return MyQVariant("%s" % node.asstr()[()][0])
+                                                except:
+                                                    return MyQVariant("%s" % node[()][0])
+                                        else:
+                                            #stored as a string
+                                            try:
+                                                try:
+                                                    return MyQVariant("%s" % node.asstr()[()])
+                                                except:
+                                                    return MyQVariant("%s" % node[()])
+                                            except:
+                                                # issue #745
+                                                return MyQVariant("Unknown %s" % node)
                                 else:
-                                    #stored as a string
-                                    return MyQVariant("%s" % node[()])
-                return MyQVariant(item.type)
-            if column == 2:
-                return MyQVariant(item.shape)
-            if column == 3:
-                return MyQVariant(item.dtype)
-        elif role == qt.Qt.ForegroundRole:
-            item = self.getProxyFromIndex(index)
-            column = index.column()
-            if column == 0:
+                                    _logger.critical("Entry %s has no children" % item.name)
+                    return MyQVariant(item.type)
+                if column == 2:
+                    return MyQVariant(item.shape)
+                if column == 3:
+                    return MyQVariant(item.dtype)
+            elif role == qt.Qt.ForegroundRole:
+                item = self.getProxyFromIndex(index)
+                column = index.column()
+                if column == 0:
+                    if hasattr(item, "color"):
+                        return MyQVariant(qt.QColor(item.color))
+            elif role == qt.Qt.ToolTipRole:
+                item = self.getProxyFromIndex(index)
                 if hasattr(item, "color"):
-                    return MyQVariant(qt.QColor(item.color))
-        elif role == qt.Qt.ToolTipRole:
-            item = self.getProxyFromIndex(index)
-            if hasattr(item, "color"):
-                if item.color == qt.Qt.blue:
-                    return MyQVariant("Item has a double click NXdata associated action")
-        return MyQVariant()
+                    if item.color == qt.Qt.blue:
+                        return MyQVariant("Item has a double click NXdata associated action")
+            return MyQVariant()
+        except:
+            return MyQVariant("Unhandled exception filling tree. Reload?")
 
     def getNodeFromIndex(self, index):
         try:
@@ -634,7 +733,6 @@ class FileView(qt.QTreeView):
             return
         self.fileUpdated(ddict)
 
-
     def fileUpdated(self, ddict):
         rootModelIndex = self.rootIndex()
         if self.model().hasChildren(rootModelIndex):
@@ -656,6 +754,8 @@ class HDF5Widget(FileView):
     def __init__(self, model, parent=None):
         FileView.__init__(self, model, parent)
         self.setSelectionBehavior(qt.QAbstractItemView.SelectRows)
+        self.setAutoScroll(False)
+
         self._adjust()
         if 0:
             self.activated[qt.QModelIndex].connect(self.itemActivated)
@@ -664,6 +764,49 @@ class HDF5Widget(FileView):
         self.doubleClicked[qt.QModelIndex].connect(self.itemDoubleClicked)
         self.collapsed[qt.QModelIndex].connect(self._adjust)
         self.expanded[qt.QModelIndex].connect(self._adjust)
+        self.setSortingEnabled(False)
+        self.header().sectionDoubleClicked[int].connect( \
+                         self._headerSectionDoubleClicked)
+        tip = "Double click on first two columns to change order"
+        self.header().setToolTip(tip)
+
+    def _headerSectionDoubleClicked(self, index):
+        self.sortItems(index, qt.Qt.AscendingOrder)
+
+    def __updateOrder(self):
+        rootModelIndex = self.rootIndex()
+        filelist = []
+        if self.model().hasChildren(rootModelIndex):
+            rootItem = self.model().getProxyFromIndex(rootModelIndex)
+            for row in range(len(rootItem)):
+                if self.model().hasIndex(row, 0, rootModelIndex):
+                    modelIndex = self.model().index(row, 0, rootModelIndex)
+                    item = self.model().getProxyFromIndex(modelIndex)
+                    try:
+                        filename = item.file.filename
+                        if filename not in filelist:
+                            filelist.append(filename)
+                    except:
+                        continue
+        if len(filelist):
+            for file in filelist:
+                ddict = {}
+                ddict['event'] = "fileUpdated"
+                ddict['filename'] = filename
+                self.fileUpdated(ddict)
+
+    def sortByColumn(self, column, order):
+        #reimplement QTreeWidget sorting
+        _logger.info("sort by column %d setting indicator %s" % (column, order))
+        self.setSortingEnabled(True)
+        self.header().setSortIndicator(column, order)
+        self.__updateOrder()
+        self.setSortingEnabled(False)
+
+    def sortItems(self, column, order):
+        #reimplement QTreeWidget sorting
+        _logger.info("sort items")
+        self.sortByColumn(column, order)
 
     def _adjust(self, modelIndex=None):
         self.resizeColumnToContents(0)
@@ -684,7 +827,7 @@ class HDF5Widget(FileView):
             self._lastMouse = "left"
         qt.QTreeView.mousePressEvent(self, e)
         if self._lastMouse != "left":
-            # Qt5 only sends itenClicked on left button mouse click
+            # Qt5 only sends itemClicked on left button mouse click
             if QVERSION > "5":
                 event = "itemClicked"
                 modelIndex = self.indexAt(e.pos())
@@ -725,7 +868,7 @@ class HDF5Widget(FileView):
         ddict['dtype'] = item.dtype
         ddict['shape'] = item.shape
         ddict['color'] = item.color
-        ddict['mouse'] = self._lastMouse * 1
+        ddict['mouse'] = getattr(self, '_lastMouse', 'left') * 1
         self.sigHDF5WidgetSignal.emit(ddict)
 
     def getSelectedEntries(self):
@@ -736,7 +879,7 @@ class HDF5Widget(FileView):
             item = self.model().getProxyFromIndex(modelIndex)
             if item.type in ["weakproxy", "File"]:
                 continue
-            filename = item.file.filename 
+            filename = item.file.filename
             path = item.name * 1
             if (path, filename) in analyzedPaths:
                 continue
@@ -847,12 +990,15 @@ class Hdf5SelectionDialog(qt.QDialog):
         self.selectedItemUri = self._lastEvent['file'] + "::" + self._lastEvent['name']
         self.accept()
 
-    def exec_(self):
+    def exec(self):
         with h5open(self.filename) as hdf5File:
             self.fileModel.appendPhynxFile(hdf5File, weakreference=True)
-            ret = qt.QDialog.exec_(self)
+            ret = qt.QDialog.exec(self)
         return ret
 
+    def exec_(self):
+        return self.exec()
+
 
 def _getFilenameDialog(parent=None):
     """Open a dialog to select a file in a filesystem tree view.
@@ -885,7 +1031,7 @@ def getDatasetValueDialog(filename=None, message=None, parent=None):
     """
     hdf5Dialog = Hdf5SelectionDialog(parent, filename, message,
                                      "dataset")
-    ret = hdf5Dialog.exec_()
+    ret = hdf5Dialog.exec()
     if not ret:
         return None
 
@@ -900,7 +1046,7 @@ def getDatasetValueDialog(filename=None, message=None, parent=None):
 def getDatasetUri(parent=None, filename=None, message=None):
     # TODO: Accept a filter for type of dataset
     hdf5Dialog = Hdf5SelectionDialog(parent, filename, message, "dataset")
-    ret = hdf5Dialog.exec_()
+    ret = hdf5Dialog.exec()
     if not ret:
         return None
     selectedHdf5Uri = hdf5Dialog.selectedItemUri
@@ -909,7 +1055,7 @@ def getDatasetUri(parent=None, filename=None, message=None):
 def getGroupUri(parent=None, filename=None, message=None):
     # TODO: Accept a filter for a particular attribute (NXclass)
     hdf5Dialog = Hdf5SelectionDialog(parent, filename, message, "dataset")
-    ret = hdf5Dialog.exec_()
+    ret = hdf5Dialog.exec()
     if not ret:
         return None
     selectedHdf5Uri = hdf5Dialog.selectedItemUri
@@ -917,7 +1063,7 @@ def getGroupUri(parent=None, filename=None, message=None):
 
 def getUri(parent=None, filename=None, message=None):
     hdf5Dialog = Hdf5SelectionDialog(parent, filename, message, "any")
-    ret = hdf5Dialog.exec_()
+    ret = hdf5Dialog.exec()
     if not ret:
         return None
     selectedHdf5Uri = hdf5Dialog.selectedItemUri
@@ -931,7 +1077,7 @@ def getDatasetDialog(filename=None, value=False, message=None, parent=None):
                                      parent=parent)
 
     hdf5Dialog = Hdf5SelectionDialog(parent, filename, message, "dataset")
-    ret = hdf5Dialog.exec_()
+    ret = hdf5Dialog.exec()
     if not ret:
         return None
     selectedHdf5Uri = hdf5Dialog.selectedItemUri
@@ -949,7 +1095,7 @@ def getGroupNameDialog(filename=None, message=None, parent=None):
     :return: HDF5 group name
     """
     hdf5Dialog = Hdf5SelectionDialog(parent, filename, message, "group")
-    ret = hdf5Dialog.exec_()
+    ret = hdf5Dialog.exec()
     if not ret:
         return None
 
@@ -973,6 +1119,6 @@ if __name__ == "__main__":
             print(phynxFile[ddict['name']].dtype, phynxFile[ddict['name']].shape)
     fileView.sigHDF5WidgetSignal.connect(mySlot)
     fileView.show()
-    ret = app.exec_()
+    ret = app.exec()
     app = None
     sys.exit(ret)
